# -*- coding: utf-8 -*-

from .grammar import SchemaGrammar
from ..blueprint import Blueprint
from ...query.expression import QueryExpression
from ...support.fluent import Fluent


class SQLiteSchemaGrammar(SchemaGrammar):

    _modifiers = ["nullable", "default", "increment"]

    _serials = ["big_integer", "integer"]

    def compile_rename_column(self, blueprint, command, connection):
        """
        Compile a rename column command.

        :param blueprint: The blueprint
        :type blueprint: Blueprint

        :param command: The command
        :type command: Fluent

        :param connection: The connection
        :type connection: orator.connections.Connection

        :rtype: list
        """
        sql = []
        # If foreign keys are on, we disable them
        foreign_keys = self._connection.select("PRAGMA foreign_keys")
        if foreign_keys:
            foreign_keys = bool(foreign_keys[0])
            if foreign_keys:
                sql.append("PRAGMA foreign_keys = OFF")

        sql += super(SQLiteSchemaGrammar, self).compile_rename_column(
            blueprint, command, connection
        )

        if foreign_keys:
            sql.append("PRAGMA foreign_keys = ON")

        return sql

    def compile_change(self, blueprint, command, connection):
        """
        Compile a change column command into a series of SQL statement.

        :param blueprint: The blueprint
        :type blueprint: orator.schema.Blueprint

        :param command: The command
        :type command: Fluent

        :param connection: The connection
        :type connection: orator.connections.Connection

        :rtype: list
        """
        sql = []
        # If foreign keys are on, we disable them
        foreign_keys = self._connection.select("PRAGMA foreign_keys")
        if foreign_keys:
            foreign_keys = bool(foreign_keys[0])
            if foreign_keys:
                sql.append("PRAGMA foreign_keys = OFF")

        sql += super(SQLiteSchemaGrammar, self).compile_change(
            blueprint, command, connection
        )

        if foreign_keys:
            sql.append("PRAGMA foreign_keys = ON")

        return sql

    def compile_table_exists(self):
        """
        Compile the query to determine if a table exists

        :rtype: str
        """
        return (
            "SELECT * FROM sqlite_master WHERE type = 'table' AND name = %(marker)s"
            % {"marker": self.get_marker()}
        )

    def compile_column_exists(self, table):
        """
        Compile the query to determine the list of columns.
        """
        return "PRAGMA table_info(%s)" % table.replace(".", "__")

    def compile_create(self, blueprint, command, _):
        """
        Compile a create table command.
        """
        columns = ", ".join(self._get_columns(blueprint))

        sql = "CREATE TABLE %s (%s" % (self.wrap_table(blueprint), columns)

        sql += self._add_foreign_keys(blueprint)

        sql += self._add_primary_keys(blueprint)

        return sql + ")"

    def _add_foreign_keys(self, blueprint):
        sql = ""

        foreigns = self._get_commands_by_name(blueprint, "foreign")

        for foreign in foreigns:
            sql += self._get_foreign_key(foreign)

            if foreign.get("on_delete"):
                sql += " ON DELETE %s" % foreign.on_delete

            if foreign.get("on_update"):
                sql += " ON UPDATE %s" % foreign.on_delete

        return sql

    def _get_foreign_key(self, foreign):
        on = self.wrap_table(foreign.on)

        columns = self.columnize(foreign.columns)

        references = foreign.references
        if not isinstance(references, list):
            references = [references]

        on_columns = self.columnize(references)

        return ", FOREIGN KEY(%s) REFERENCES %s(%s)" % (columns, on, on_columns)

    def _add_primary_keys(self, blueprint):
        primary = self._get_command_by_name(blueprint, "primary")

        if primary:
            columns = self.columnize(primary.columns)

            return ", PRIMARY KEY (%s)" % columns

        return ""

    def compile_add(self, blueprint, command, _):
        table = self.wrap_table(blueprint)

        columns = self.prefix_list("ADD COLUMN", self._get_columns(blueprint))

        statements = []

        for column in columns:
            statements.append("ALTER TABLE %s %s" % (table, column))

        return statements

    def compile_unique(self, blueprint, command, _):
        columns = self.columnize(command.columns)

        table = self.wrap_table(blueprint)

        return "CREATE UNIQUE INDEX %s ON %s (%s)" % (command.index, table, columns)

    def compile_index(self, blueprint, command, _):
        columns = self.columnize(command.columns)

        table = self.wrap_table(blueprint)

        return "CREATE INDEX %s ON %s (%s)" % (command.index, table, columns)

    def compile_foreign(self, blueprint, command, _):
        pass

    def compile_drop(self, blueprint, command, _):
        return "DROP TABLE %s" % self.wrap_table(blueprint)

    def compile_drop_if_exists(self, blueprint, command, _):
        return "DROP TABLE IF EXISTS %s" % self.wrap_table(blueprint)

    def compile_drop_column(self, blueprint, command, connection):
        schema = connection.get_schema_manager()

        table_diff = self._get_table_diff(blueprint, schema)

        for name in command.columns:
            column = connection.get_column(blueprint.get_table(), name)

            table_diff.removed_columns[name] = column

        return schema.get_database_platform().get_alter_table_sql(table_diff)

    def compile_drop_unique(self, blueprint, command, _):
        return "DROP INDEX %s" % command.index

    def compile_drop_index(self, blueprint, command, _):
        return "DROP INDEX %s" % command.index

    def compile_rename(self, blueprint, command, _):
        from_ = self.wrap_table(blueprint)

        return "ALTER TABLE %s RENAME TO %s" % (from_, self.wrap_table(command.to))

    def _type_char(self, column):
        return "VARCHAR"

    def _type_string(self, column):
        return "VARCHAR"

    def _type_text(self, column):
        return "TEXT"

    def _type_medium_text(self, column):
        return "TEXT"

    def _type_long_text(self, column):
        return "TEXT"

    def _type_integer(self, column):
        return "INTEGER"

    def _type_big_integer(self, column):
        return "INTEGER"

    def _type_medium_integer(self, column):
        return "INTEGER"

    def _type_tiny_integer(self, column):
        return "TINYINT"

    def _type_small_integer(self, column):
        return "INTEGER"

    def _type_float(self, column):
        return "FLOAT"

    def _type_double(self, column):
        return "FLOAT"

    def _type_decimal(self, column):
        return "NUMERIC"

    def _type_boolean(self, column):
        return "TINYINT"

    def _type_enum(self, column):
        return "VARCHAR"

    def _type_json(self, column):
        return "TEXT"

    def _type_date(self, column):
        return "DATE"

    def _type_datetime(self, column):
        return "DATETIME"

    def _type_time(self, column):
        return "TIME"

    def _type_timestamp(self, column):
        if column.use_current:
            return "DATETIME DEFAULT CURRENT_TIMESTAMP"

        return "DATETIME"

    def _type_binary(self, column):
        return "BLOB"

    def _modify_nullable(self, blueprint, column):
        if column.get("nullable"):
            return " NULL"

        return " NOT NULL"

    def _modify_default(self, blueprint, column):
        if column.get("default") is not None:
            return " DEFAULT %s" % self._get_default_value(column.default)

        return ""

    def _modify_increment(self, blueprint, column):
        if column.type in self._serials and column.auto_increment:
            return " PRIMARY KEY AUTOINCREMENT"

        return ""

    def _get_dbal_column_type(self, type_):
        """
        Get the dbal column type.

        :param type_: The fluent type
        :type type_: str

        :rtype: str
        """
        type_ = type_.lower()

        if type_ == "enum":
            return "string"

        return super(SQLiteSchemaGrammar, self)._get_dbal_column_type(type_)



# -*- coding: utf-8 -*-

import os
from orbit_orator import DatabaseManager
from cleo import Command as BaseCommand, InputOption, ListInput
import yaml


class Command(BaseCommand):

    needs_config = True

    def __init__(self):
        super(Command, self).__init__()

        self._input = None
        self._output = None

    def configure(self):
        super(Command, self).configure()

        if self.needs_config and not self.resolver:
            # Checking if a default config file is present
            if not self._check_config():
                self.add_option(
                    "config", "c", InputOption.VALUE_REQUIRED, "The config file path"
                )

    def execute(self, i, o):
        """
        Executes the command.
        """
        self.set_style("question", fg="blue")

        if self.needs_config and not self.resolver:
            self._handle_config(self.option("config"))

        return self.handle()

    def call(self, name, options=None):
        command = self.get_application().find(name)
        command.resolver = self.resolver

        return super(Command, self).call(name, options)

    def call_silent(self, name, options=None):
        command = self.get_application().find(name)
        command.resolver = self.resolver

        return super(Command, self).call_silent(name, options)

    def confirm_to_proceed(self, message=None):
        if message is None:
            message = "Do you really wish to run this command?: "

        if self.option("force"):
            return True

        confirmed = self.confirm(message)

        if not confirmed:
            self.comment("Command Cancelled!")

            return False

        return True

    def _get_migration_path(self):
        return os.path.join(os.getcwd(), "migrations")

    def _check_config(self):
        """
        Check presence of default config files.

        :rtype: bool
        """
        current_path = os.path.relpath(os.getcwd())

        accepted_files = ["orator.yml", "orator.py"]
        for accepted_file in accepted_files:
            config_file = os.path.join(current_path, accepted_file)
            if os.path.exists(config_file):
                if self._handle_config(config_file):
                    return True

        return False

    def _handle_config(self, config_file):
        """
        Check and handle a config file.

        :param config_file: The path to the config file
        :type config_file: str

        :rtype: bool
        """
        config = self._get_config(config_file)

        self.resolver = DatabaseManager(
            config.get("databases", config.get("DATABASES", {}))
        )

        return True

    def _get_config(self, path=None):
        """
        Get the config.

        :rtype: dict
        """
        if not path and not self.option("config"):
            raise Exception("The --config|-c option is missing.")

        if not path:
            path = self.option("config")

        filename, ext = os.path.splitext(path)
        if ext in [".yml", ".yaml"]:
            with open(path) as fd:
                config = yaml.load(fd)
        elif ext in [".py"]:
            config = {}

            with open(path) as fh:
                exec(fh.read(), {}, config)
        else:
            raise RuntimeError("Config file [%s] is not supported." % path)

        return config



# -*- coding: utf-8 -*-

import re
from pendulum import DateTime, Date

try:
    import MySQLdb as mysql

    # Fix for understanding DateTime object
    import MySQLdb.converters

    MySQLdb.converters.conversions[DateTime] = MySQLdb.converters.DateTime2literal
    MySQLdb.converters.conversions[Date] = MySQLdb.converters.Thing2Literal

    from MySQLdb.cursors import DictCursor as cursor_class

    keys_fix = {"password": "passwd", "database": "db"}
except ImportError as e:
    try:
        import pymysql as mysql

        # Fix for understanding DateTime object
        import pymysql.converters

        pymysql.converters.conversions[DateTime] = pymysql.converters.escape_datetime
        pymysql.converters.conversions[Date] = pymysql.converters.escape_date

        from pymysql.cursors import DictCursor as cursor_class

        keys_fix = {}
    except ImportError as e:
        mysql = None
        cursor_class = object

from ..dbal.platforms import MySQLPlatform, MySQL57Platform
from .connector import Connector
from ..utils.qmarker import qmark, denullify
from ..utils.helpers import serialize


class Record(dict):
    def __getattr__(self, item):
        try:
            return self[item]
        except KeyError:
            raise AttributeError(item)

    def serialize(self):
        return serialize(self)


class BaseDictCursor(cursor_class):
    def _fetch_row(self, size=1):
        # Overridden for mysqclient
        if not self._result:
            return ()
        rows = self._result.fetch_row(size, self._fetch_type)

        return tuple(Record(r) for r in rows)

    def _conv_row(self, row):
        # Overridden for pymysql
        return Record(super(BaseDictCursor, self)._conv_row(row))


class DictCursor(BaseDictCursor):
    def execute(self, query, args=None):
        query = qmark(query)

        return super(DictCursor, self).execute(query, args)

    def executemany(self, query, args):
        query = qmark(query)

        return super(DictCursor, self).executemany(query, denullify(args))


class MySQLConnector(Connector):

    RESERVED_KEYWORDS = [
        "log_queries",
        "driver",
        "prefix",
        "engine",
        "collation",
        "name",
        "use_qmark",
    ]

    SUPPORTED_PACKAGES = ["PyMySQL", "mysqlclient"]

    def _do_connect(self, config):
        config = dict(config.items())
        for key, value in keys_fix.items():
            config[value] = config[key]
            del config[key]

        config["autocommit"] = True
        config["cursorclass"] = self.get_cursor_class(config)

        return self.get_api().connect(**self.get_config(config))

    def get_default_config(self):
        return {"charset": "utf8", "use_unicode": True}

    def get_cursor_class(self, config):
        if config.get("use_qmark"):
            return DictCursor

        return BaseDictCursor

    def get_api(self):
        return mysql

    def get_server_version(self):
        version = self._connection.get_server_info()

        version_parts = re.match(
            "^(?P<major>\d+)(?:\.(?P<minor>\d+)(?:\.(?P<patch>\d+))?)?", version
        )

        major = int(version_parts.group("major"))
        minor = version_parts.group("minor") or 0
        patch = version_parts.group("patch") or 0

        minor, patch = int(minor), int(patch)

        server_version = (major, minor, patch, "")

        if "mariadb" in version.lower():
            server_version = (major, minor, patch, "mariadb")

        return server_version

    def _create_database_platform_for_version(self, version):
        major, minor, _, extra = version

        if extra == "mariadb":
            return self.get_dbal_platform()

        if (major, minor) >= (5, 7):
            return MySQL57Platform()

        return self.get_dbal_platform()

    def get_dbal_platform(self):
        return MySQLPlatform()



# -*- coding: utf-8 -*-

from .base_command import BaseCommand


class RefreshCommand(BaseCommand):
    """
    Reset and re-run all migrations.

    migrate:refresh
        {--d|database= : The database connection to use.}
        {--p|path= : The path of migrations files to be executed.}
        {--s|seed : Indicates if the seed task should be re-run.}
        {--seed-path= : The path of seeds files to be executed.
                        Defaults to <comment>./seeds</comment>.}
        {--seeder=database_seeder : The name of the root seeder.}
        {--f|force : Force the operation to run.}
    """

    def handle(self):
        """
        Executes the command.
        """
        if not self.confirm_to_proceed(
            "<question>Are you sure you want to refresh the database?:</question> "
        ):
            return

        database = self.option("database")

        options = [("--force", True)]

        if self.option("path"):
            options.append(("--path", self.option("path")))

        if database:
            options.append(("--database", database))

        if self.get_definition().has_option("config"):
            options.append(("--config", self.option("config")))

        self.call("migrate:reset", options)

        self.call("migrate", options)

        if self._needs_seeding():
            self._run_seeder(database)

    def _needs_seeding(self):
        return self.option("seed")

    def _run_seeder(self, database):
        options = [("--seeder", self.option("seeder")), ("--force", True)]

        if database:
            options.append(("--database", database))

        if self.get_definition().has_option("config"):
            options.append(("--config", self.option("config")))

        if self.option("seed-path"):
            options.append(("--path", self.option("seed-path")))

        self.call("db:seed", options)



# -*- coding: utf-8 -*-

from orbit_orator.migrations import DatabaseMigrationRepository
from .base_command import BaseCommand


class InstallCommand(BaseCommand):
    """
    Install the migration repository.

    migrate:install
        {--d|database= : The database connection to use.}
    """

    def handle(self):
        """
        Execute the console command.
        """
        database = self.option("database")
        repository = DatabaseMigrationRepository(self.resolver, "migrations")

        repository.set_source(database)
        repository.create_repository()

        self.info("Migration table created successfully")



# -*- coding: utf-8 -*-

from orbit_orator.migrations import Migrator, DatabaseMigrationRepository
from .base_command import BaseCommand


class StatusCommand(BaseCommand):
    """
    Show the status of each migration.

    migrate:status
        {--d|database= : The database connection to use.}
        {--p|path= : The path of migrations files to be executed.}
    """

    def handle(self):
        """
        Execute the console command.
        """
        database = self.option('database')

        repository = DatabaseMigrationRepository(self.resolver, self.config.get('migrations.table', 'migrations'))

        if not repository.repository_exists():
            return self.error('No migrations found.')

        migrator = Migrator(repository, self.resolver)

        ran = migrator.get_repository().get_ran()

        migrations = []
        for migration in migrator.get_migrations_files(self._get_migration_path()):
            migrations.append({
                'migration': migration,
                'batch': migrator.get_repository().get_last_batch_number() if migration in ran else '',
                'ran': 'Yes' if migration in ran else 'No'
            })

        if not migrations:
            return self.error('No migrations found')

        self.render_status(migrations, database)

    def _prepare_database(self, migrator, database):
        migrator.set_connection(database)



# -*- coding: utf-8 -*-

import os
import errno
import datetime


def value(val):
    if callable(val):
        return val()

    return val


def mkdir_p(path, mode=0o777):
    try:
        os.makedirs(path, mode)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def serialize(value):
    if isinstance(value, datetime.datetime):
        if hasattr(value, "to_json"):
            value = value.to_json()
        else:
            value = value.isoformat()
    elif isinstance(value, list):
        value = list(map(serialize, value))
    elif isinstance(value, dict):
        for k, v in value.items():
            value[k] = serialize(v)

    return value



"""
Orator - A simple ORM for Python
"""

__version__ = "0.9.9.11"

from orbit_orator.orm import Model
from orbit_orator.orm.relations import (
    HasOne,
    HasMany,
    BelongsTo,
    BelongsToMany,
    MorphOne,
    MorphMany,
    MorphTo,
    MorphToMany,
    HasManyThrough,
)
from orbit_orator.query.builder import QueryBuilder
from orbit_orator.schema.builder import SchemaBuilder
from orbit_orator.connections import Connection
from orbit_orator.database_manager import DatabaseManager

__all__ = [
    "Model",
    "HasOne",
    "HasMany",
    "BelongsTo",
    "BelongsToMany",
    "MorphOne",
    "MorphMany",
    "MorphTo",
    "MorphToMany",
    "HasManyThrough",
    "QueryBuilder",
    "SchemaBuilder",
    "Connection",
    "DatabaseManager",
]



# -*- coding: utf-8 -*-

from .grammar import QueryGrammar
from ...utils import basestring


class PostgresQueryGrammar(QueryGrammar):

    _operators = [
        "=",
        "<",
        ">",
        "<=",
        ">=",
        "<>",
        "!=",
        "like",
        "not like",
        "between",
        "ilike",
        "&",
        "|",
        "#",
        "<<",
        ">>",
    ]

    marker = "%s"

    def _compile_lock(self, query, value):
        """
        Compile the lock into SQL

        :param query: A QueryBuilder instance
        :type query: QueryBuilder

        :param value: The lock value
        :type value: bool or str

        :return: The compiled lock
        :rtype: str
        """
        if isinstance(value, basestring):
            return value

        if value:
            return "FOR UPDATE"

        return "FOR SHARE"

    def compile_update(self, query, values):
        """
        Compile an update statement into SQL

        :param query: A QueryBuilder instance
        :type query: QueryBuilder

        :param values: The update values
        :type values: dict

        :return: The compiled update
        :rtype: str
        """
        table = self.wrap_table(query.from__)

        columns = self._compile_update_columns(values)

        from_ = self._compile_update_from(query)

        where = self._compile_update_wheres(query)

        return ("UPDATE %s SET %s%s %s" % (table, columns, from_, where)).strip()

    def _compile_update_columns(self, values):
        """
        Compile the columns for the update statement

        :param values: The columns
        :type values: dict

        :return: The compiled columns
        :rtype: str
        """
        columns = []

        for key, value in values.items():
            columns.append("%s = %s" % (self.wrap(key), self.parameter(value)))

        return ", ".join(columns)

    def _compile_update_from(self, query):
        """
        Compile the "from" clause for an update with a join.

        :param query: A QueryBuilder instance
        :type query: QueryBuilder

        :return: The compiled sql
        :rtype: str
        """
        if not query.joins:
            return ""

        froms = []

        for join in query.joins:
            froms.append(self.wrap_table(join.table))

        if len(froms):
            return " FROM %s" % ", ".join(froms)

        return ""

    def _compile_update_wheres(self, query):
        """
        Compile the additional where clauses for updates with joins.

        :param query: A QueryBuilder instance
        :type query: QueryBuilder

        :return: The compiled sql
        :rtype: str
        """
        base_where = self._compile_wheres(query)

        if not query.joins:
            return base_where

        join_where = self._compile_update_join_wheres(query)

        if not base_where.strip():
            return "WHERE %s" % self._remove_leading_boolean(join_where)

        return "%s %s" % (base_where, join_where)

    def _compile_update_join_wheres(self, query):
        """
        Compile the "join" clauses for an update.

        :param query: A QueryBuilder instance
        :type query: QueryBuilder

        :return: The compiled sql
        :rtype: str
        """
        join_wheres = []

        for join in query.joins:
            for clause in join.clauses:
                join_wheres.append(self._compile_join_constraints(clause))

        return " ".join(join_wheres)

    def compile_insert_get_id(self, query, values, sequence=None):
        """
        Compile an insert and get ID statement into SQL.

        :param query: A QueryBuilder instance
        :type query: QueryBuilder

        :param values: The values to insert
        :type values: dict

        :param sequence: The id sequence
        :type sequence: str

        :return: The compiled statement
        :rtype: str
        """
        if sequence is None:
            sequence = "id"

        return "%s RETURNING %s" % (
            self.compile_insert(query, values),
            self.wrap(sequence),
        )

    def compile_truncate(self, query):
        """
        Compile a truncate table statement into SQL.

        :param query: A QueryBuilder instance
        :type query: QueryBuilder

        :return: The compiled statement
        :rtype: str
        """
        return {"TRUNCATE %s RESTART IDENTITY" % self.wrap_table(query.from__): {}}



# -*- coding: utf-8 -*-

from .processor import QueryProcessor
from .mysql_processor import MySQLQueryProcessor
from .postgres_processor import PostgresQueryProcessor
from .sqlite_processor import SQLiteQueryProcessor



# -*- coding: utf-8 -*-

from ..dbal.exceptions import InvalidPlatformSpecified
from ..exceptions.connectors import MissingPackage


class Connector(object):

    RESERVED_KEYWORDS = ["log_queries", "driver", "prefix", "name"]

    SUPPORTED_PACKAGES = []

    def __init__(self, driver=None):
        if self.get_api() is None:
            raise MissingPackage(driver, self.SUPPORTED_PACKAGES)

        self._connection = None
        self._platform = None
        self._params = {}

    def get_api(self):
        raise NotImplementedError()

    def get_config(self, config):
        default_config = self.get_default_config()
        config = {x: config[x] for x in config if x not in self.RESERVED_KEYWORDS}

        default_config.update(config)

        return default_config

    def get_default_config(self):
        return {}

    def connect(self, config):
        self._params = self.get_config(config)
        self._connection = self._do_connect(config)

        return self

    def _do_connect(self, config):
        return self.get_api().connect(**self.get_config(config))

    def get_params(self):
        return self._params

    def get_database(self):
        return self._params.get("database")

    def get_host(self):
        return self._params.get("host")

    def get_user(self):
        return self._params.get("user")

    def get_password(self):
        return self._params.get("password")

    def get_database_platform(self):
        if self._platform is None:
            self._detect_database_platform()

        return self._platform

    def _detect_database_platform(self):
        """
        Detects and sets the database platform.

        Evaluates custom platform class and version in order to set the correct platform.

        :raises InvalidPlatformSpecified: if an invalid platform was specified for this connection.
        """
        version = self._get_database_platform_version()

        if version is not None:
            self._platform = self._create_database_platform_for_version(version)
        else:
            self._platform = self.get_dbal_platform()

    def _get_database_platform_version(self):
        """
        Returns the version of the related platform if applicable.

        Returns None if either the connector is not capable to create version
        specific platform instances, no explicit server version was specified
        or the underlying driver connection cannot determine the platform
        version without having to query it (performance reasons).

        :rtype: str or None
        """
        # Connector does not support version specific platforms.
        if not self.is_version_aware():
            return None

        return self.get_server_version()

    def _create_database_platform_for_version(self, version):
        raise NotImplementedError()

    def get_dbal_platform(self):
        raise NotImplementedError()

    def is_version_aware(self):
        return True

    def get_server_version(self):
        return None

    def __getattr__(self, item):
        return getattr(self._connection, item)



# -*- coding: utf-8 -*-

import os
from orbit_orator.migrations import MigrationCreator
from .base_command import BaseCommand


class MigrateMakeCommand(BaseCommand):
    """
    Create a new migration file.

    make:migration
        {name : The name of the migration.}
        {--t|table= : The table to create the migration for.}
        {--C|create : Whether the migration will create the table or not.}
        {--p|path= : The path to migrations files.}
    """

    needs_config = False

    def handle(self):
        """
        Execute the console command.
        """
        creator = MigrationCreator()

        name = self.argument('name')
        table = self.option('table')
        create = bool(self.option('create'))

        if not table and create:
            table = create

        path = creator.create(name, os.path.join(os.getcwd(), 'migrations'), table, create)

        self.line('<info>Created migration:</info> %s' % os.path.basename(path))

    def _write_migration(self, creator, name, table, create, path):
        """
        Write the migration file to disk.
        """
        file_ = os.path.basename(creator.create(name, path, table, create))

        return file_



# -*- coding: utf-8 -*-

DEFAULT_STUB = """from orbit_orator.seeds import Seeder


class DummyClass(Seeder):

    def run(self):
        \"\"\"
        Run the database seeds.
        \"\"\"
        pass

"""



# -*- coding: utf-8 -*-

from . import OrbitOratorTestCase
from . import mock
from .utils import MockConnection, MockManager

from orbit_orator.database_manager import DatabaseManager


class ConnectionTestCase(OrbitOratorTestCase):
    def test_connection_method_create_a_new_connection_if_needed(self):
        manager = self._get_manager()
        manager.table("users")

        manager._make_connection.assert_called_once_with("sqlite")

        manager._make_connection.reset_mock()
        manager.table("users")
        self.assertFalse(manager._make_connection.called)

    def test_manager_uses_factory_to_create_connections(self):
        manager = self._get_real_manager()
        original_make = manager._factory.make
        manager._factory.make = mock.MagicMock()
        manager.connection()

        manager._factory.make.assert_called_with(
            {"name": "sqlite", "driver": "sqlite", "database": ":memory:"}, "sqlite"
        )

        manager._factory.make = original_make

    def test_connection_can_select_connections(self):
        manager = self._get_manager()
        self.assertEqual(manager.connection(), manager.connection("sqlite"))
        self.assertNotEqual(manager.connection("sqlite"), manager.connection("sqlite2"))

    def test_dynamic_attribute_gets_connection_attribute(self):
        manager = self._get_manager()
        manager.statement("CREATE TABLE users")

        manager.get_connections()["sqlite"].statement.assert_called_once_with(
            "CREATE TABLE users"
        )

    def test_default_database_with_one_database(self):
        manager = MockManager(
            {"sqlite": {"driver": "sqlite", "database": ":memory:"}}
        ).prepare_mock()

        self.assertEqual("sqlite", manager.get_default_connection())

    def test_reconnect(self):
        manager = self._get_real_manager()

        api_connection = manager.connection().get_connection()
        manager.reconnect()
        self.assertIsNot(manager.connection().get_connection(), api_connection)
        self.assertIsNotNone(manager.connection().get_connection())

    def test_set_default_connection_with_none_should_not_overwrite(self):
        manager = self._get_real_manager()

        manager.set_default_connection(None)

        self.assertEqual("sqlite", manager.get_default_connection())

    def _get_manager(self):
        manager = MockManager(
            {
                "default": "sqlite",
                "sqlite": {"driver": "sqlite", "database": ":memory:"},
                "sqlite2": {"driver": "sqlite", "database": ":memory:"},
            }
        ).prepare_mock()

        return manager

    def _get_real_manager(self):
        manager = DatabaseManager(
            {
                "default": "sqlite",
                "sqlite": {"driver": "sqlite", "database": ":memory:"},
            }
        )

        return manager

    def _get_connection(self):
        return MockConnection().prepare_mock()



# -*- coding: utf-8 -*-

from orbit_orator.migrations import Migrator, DatabaseMigrationRepository
from .base_command import BaseCommand


class MigrateCommand(BaseCommand):
    """
    Run the database migrations.

    migrate
        {--d|database= : The database connection to use.}
        {--p|path= : The path of migrations files to be executed.}
        {--s|seed : Indicates if the seed task should be re-run.}
        {--seed-path= : The path of seeds files to be executed.
                        Defaults to <comment>./seeders</comment>.}
        {--P|pretend : Dump the SQL queries that would be run.}
        {--f|force : Force the operation to run.}
    """

    def handle(self):
        """
        Execute the console command.
        """
        if not self.confirm_to_proceed(
            "<question>Are you sure you want to proceed with the migration?</question> "
        ):
            return

        self._prepare_database()

        pretend = bool(self.option('pretend'))
        step = self.option('step')

        if step and not isinstance(step, bool):
            step = int(step)

        database = self.option('database')

        repository = DatabaseMigrationRepository(self.resolver, self.config.get('migrations.table', 'migrations'))

        migrator = Migrator(repository, self.resolver)

        if not self.option('pretend'):
            migrator.run(self._get_migration_path(), pretend)

            for note in migrator.get_notes():
                self.line(note)

            if step > 0:
                migrator.rollback(self._get_migration_path(), step)

                for note in migrator.get_notes():
                    self.line(note)

            if database:
                self.line('<info>Database %s was migrated.</info>' % database)
            else:
                self.line('<info>Database was migrated.</info>')

        # If the "seed" option has been given, we will rerun the database seed task
        # to repopulate the database.
        if self.option("seed"):
            options = [("--force", self.option("force"))]

            if database:
                options.append(("--database", database))

            if self.get_definition().has_option("config"):
                options.append(("--config", self.option("config")))

            if self.option("seed-path"):
                options.append(("--path", self.option("seed-path")))

            self.call("db:seed", options)

    def _prepare_database(self):
        """
        Prepare the migration database for running.
        """
        database = self.option('database')

        repository = DatabaseMigrationRepository(self.resolver, self.config.get('migrations.table', 'migrations'))

        if not repository.repository_exists():
            repository.create_repository()



# -*- coding: utf-8 -*-

import copy
from collections import OrderedDict
from ..exceptions.orm import ModelNotFound
from ..utils import Null, basestring
from ..query.expression import QueryExpression
from ..pagination import Paginator, LengthAwarePaginator
from ..support import Collection
from .scopes import Scope


class Builder(object):

    _passthru = [
        "to_sql",
        "lists",
        "insert",
        "insert_get_id",
        "pluck",
        "count",
        "min",
        "max",
        "avg",
        "sum",
        "exists",
        "get_bindings",
        "raw",
    ]

    def __init__(self, query):
        """
        Constructor

        :param query: The underlying query builder
        :type query: QueryBuilder
        """
        self._query = query

        self._model = None
        self._eager_load = {}
        self._macros = {}
        self._scopes = OrderedDict()

        self._on_delete = None

    def with_global_scope(self, identifier, scope):
        """
        Register a new global scope.

        :param identifier: The scope's identifier
        :type identifier: str

        :param scope: The scope to register
        :type scope: Scope or callable

        :rtype: Builder
        """
        self._scopes[identifier] = scope

        return self

    def without_global_scope(self, scope):
        """
        Remove a registered global scope.

        :param scope: The scope to remove
        :type scope: Scope or str

        :rtype: Builder
        """
        if isinstance(scope, basestring):
            del self._scopes[scope]

            return self

        keys = []
        for key, value in self._scopes.items():
            if scope == value.__class__ or isinstance(scope, value.__class__):
                keys.append(key)

        for key in keys:
            del self._scopes[key]

        return self

    def without_global_scopes(self):
        """
        Remove all registered global scopes.

        :rtype: Builder
        """
        self._scopes = OrderedDict()

        return self

    def find(self, id, columns=None):
        """
        Find a model by its primary key

        :param id: The primary key value
        :type id: mixed

        :param columns: The columns to retrieve
        :type columns: list

        :return: The found model
        :rtype: orator.Model
        """
        if columns is None:
            columns = ["*"]

        if isinstance(id, list):
            return self.find_many(id, columns)

        self._query.where(self._model.get_qualified_key_name(), "=", id)

        return self.first(columns)

    def find_many(self, id, columns=None):
        """
        Find a model by its primary key

        :param id: The primary key values
        :type id: list

        :param columns: The columns to retrieve
        :type columns: list

        :return: The found model
        :rtype: orator.Collection
        """
        if columns is None:
            columns = ["*"]

        if not id:
            return self._model.new_collection()

        self._query.where_in(self._model.get_qualified_key_name(), id)

        return self.get(columns)

    def find_or_fail(self, id, columns=None):
        """
        Find a model by its primary key or raise an exception

        :param id: The primary key value
        :type id: mixed

        :param columns: The columns to retrieve
        :type columns: list

        :return: The found model
        :rtype: orator.Model

        :raises: ModelNotFound
        """
        result = self.find(id, columns)

        if isinstance(id, list):
            if len(result) == len(set(id)):
                return result
        elif result:
            return result

        raise ModelNotFound(self._model.__class__)

    def first(self, columns=None):
        """
        Execute the query and get the first result

        :param columns: The columns to get
        :type columns: list

        :return: The result
        :rtype: mixed
        """
        if columns is None:
            columns = ["*"]

        return self.take(1).get(columns).first()

    def first_or_fail(self, columns=None):
        """
        Execute the query and get the first result or raise an exception

        :param columns: The columns to get
        :type columns: list

        :return: The result
        :rtype: mixed
        """
        model = self.first(columns)

        if model is not None:
            return model

        raise ModelNotFound(self._model.__class__)

    def get(self, columns=None):
        """
        Execute the query as a "select" statement.

        :param columns: The columns to get
        :type columns: list

        :rtype: orator.Collection
        """
        models = self.get_models(columns)

        # If we actually found models we will also eager load any relationships that
        # have been specified as needing to be eager loaded, which will solve the
        # n+1 query issue for the developers to avoid running a lot of queries.
        if len(models) > 0:
            models = self.eager_load_relations(models)

        collection = self._model.new_collection(models)

        return collection

    def pluck(self, column):
        """
        Pluck a single column from the database.

        :param column: THe column to pluck
        :type column: str

        :return: The column value
        :rtype: mixed
        """
        result = self.first([column])

        if result:
            return result[column]

    def chunk(self, count):
        """
        Chunk the results of the query

        :param count: The chunk size
        :type count: int

        :return: The current chunk
        :rtype: list
        """
        connection = self._model.get_connection_name()
        for results in self.apply_scopes().get_query().chunk(count):
            models = self._model.hydrate(results, connection)

            # If we actually found models we will also eager load any relationships that
            # have been specified as needing to be eager loaded, which will solve the
            # n+1 query issue for the developers to avoid running a lot of queries.
            if len(models) > 0:
                models = self.eager_load_relations(models)

            collection = self._model.new_collection(models)

            yield collection

    def lists(self, column, key=None):
        """
        Get a list with the values of a given column

        :param column: The column to get the values for
        :type column: str

        :param key: The key
        :type key: str

        :return: The list of values
        :rtype: list or dict
        """
        results = self.to_base().lists(column, key)

        if not self._model.has_get_mutator(column):
            return results

        if isinstance(results, dict):
            for key, value in results.items():
                fill = {column: value}

                results[key] = self._model.new_from_builder(fill).column
        else:
            for i, value in enumerate(results):
                fill = {column: value}

                results[i] = self._model.new_from_builder(fill).column

    def paginate(self, per_page=None, current_page=None, columns=None):
        """
        Paginate the given query.

        :param per_page: The number of records per page
        :type per_page: int

        :param current_page: The current page of results
        :type current_page: int

        :param columns: The columns to return
        :type columns: list

        :return: The paginator
        """
        if columns is None:
            columns = ["*"]

        total = self.to_base().get_count_for_pagination()

        page = current_page or Paginator.resolve_current_page()
        per_page = per_page or self._model.get_per_page()
        self._query.for_page(page, per_page)

        return LengthAwarePaginator(self.get(columns).all(), total, per_page, page)

    def simple_paginate(self, per_page=None, current_page=None, columns=None):
        """
        Paginate the given query.

        :param per_page: The number of records per page
        :type per_page: int

        :param current_page: The current page of results
        :type current_page: int

        :param columns: The columns to return
        :type columns: list

        :return: The paginator
        """
        if columns is None:
            columns = ["*"]

        page = current_page or Paginator.resolve_current_page()
        per_page = per_page or self._model.get_per_page()

        self.skip((page - 1) * per_page).take(per_page + 1)

        return Paginator(self.get(columns).all(), per_page, page)

    def update(self, _values=None, **values):
        """
        Update a record in the database

        :param values: The values of the update
        :type values: dict

        :return: The number of records affected
        :rtype: int
        """
        if _values is not None:
            values.update(_values)

        return self._query.update(self._add_updated_at_column(values))

    def increment(self, column, amount=1, extras=None):
        """
        Increment a column's value by a given amount

        :param column: The column to increment
        :type column: str

        :param amount: The amount by which to increment
        :type amount: int

        :param extras: Extra columns
        :type extras: dict

        :return: The number of rows affected
        :rtype: int
        """
        if extras is None:
            extras = {}

        extras = self._add_updated_at_column(extras)

        return self.to_base().increment(column, amount, extras)

    def decrement(self, column, amount=1, extras=None):
        """
        Decrement a column's value by a given amount

        :param column: The column to increment
        :type column: str

        :param amount: The amount by which to increment
        :type amount: int

        :param extras: Extra columns
        :type extras: dict

        :return: The number of rows affected
        :rtype: int
        """
        if extras is None:
            extras = {}

        extras = self._add_updated_at_column(extras)

        return self.to_base().decrement(column, amount, extras)

    def _add_updated_at_column(self, values):
        """
        Add the "updated_at" column to a dictionary of values.

        :param values: The values to update
        :type values: dict

        :return: The new dictionary of values
        :rtype: dict
        """
        if not self._model.uses_timestamps():
            return values

        column = self._model.get_updated_at_column()

        if "updated_at" not in values:
            values.update({column: self._model.fresh_timestamp_string()})

        return values

    def delete(self):
        """
        Delete a record from the database.
        """
        if self._on_delete is not None:
            return self._on_delete(self)

        return self._query.delete()

    def force_delete(self):
        """
        Run the default delete function on the builder.
        """
        return self._query.delete()

    def on_delete(self, callback):
        """
        Register a replacement for the default delete function.

        :param callback: A replacement for the default delete function
        :type callback: callable
        """
        self._on_delete = callback

    def get_models(self, columns=None):
        """
        Get the hydrated models without eager loading.

        :param columns: The columns to get
        :type columns: list

        :return: A list of models
        :rtype: orator.orm.collection.Collection
        """
        results = self.apply_scopes().get_query().get(columns).all()

        connection = self._model.get_connection_name()

        models = self._model.hydrate(results, connection)

        return models

    def eager_load_relations(self, models):
        """
        Eager load the relationship of the models.

        :param models:
        :type models: list

        :return: The models
        :rtype: list
        """
        for name, constraints in self._eager_load.items():
            if name.find(".") == -1:
                models = self._load_relation(models, name, constraints)

        return models

    def _load_relation(self, models, name, constraints):
        """
        Eagerly load the relationship on a set of models.

        :rtype: list
        """
        relation = self.get_relation(name)

        relation.add_eager_constraints(models)

        if callable(constraints):
            constraints(relation.get_query())
        else:
            relation.merge_query(constraints)

        models = relation.init_relation(models, name)

        results = relation.get_eager()

        return relation.match(models, results, name)

    def get_relation(self, relation):
        """
        Get the relation instance for the given relation name.

        :rtype: orator.orm.relations.Relation
        """
        from .relations import Relation

        with Relation.no_constraints(True):
            rel = getattr(self.get_model(), relation)()

        nested = self._nested_relations(relation)

        if len(nested) > 0:
            rel.get_query().with_(nested)

        return rel

    def _nested_relations(self, relation):
        """
        Get the deeply nested relations for a given top-level relation.

        :rtype: dict
        """
        nested = {}

        for name, constraints in self._eager_load.items():
            if self._is_nested(name, relation):
                nested[name[len(relation + ".") :]] = constraints

        return nested

    def _is_nested(self, name, relation):
        """
        Determine if the relationship is nested.

        :type name: str
        :type relation: str

        :rtype: bool
        """
        dots = name.find(".")

        return dots and name.startswith(relation + ".")

    def where(self, column, operator=Null(), value=None, boolean="and"):
        """
        Add a where clause to the query

        :param column: The column of the where clause, can also be a QueryBuilder instance for sub where
        :type column: str|Builder

        :param operator: The operator of the where clause
        :type operator: str

        :param value: The value of the where clause
        :type value: mixed

        :param boolean: The boolean of the where clause
        :type boolean: str

        :return: The current Builder instance
        :rtype: Builder
        """
        if isinstance(column, Builder):
            self._query.add_nested_where_query(column.get_query(), boolean)
        else:
            self._query.where(column, operator, value, boolean)

        return self

    def or_where(self, column, operator=None, value=None):
        """
        Add an "or where" clause to the query.

        :param column: The column of the where clause, can also be a QueryBuilder instance for sub where
        :type column: str or Builder

        :param operator: The operator of the where clause
        :type operator: str

        :param value: The value of the where clause
        :type value: mixed

        :return: The current Builder instance
        :rtype: Builder
        """
        return self.where(column, operator, value, "or")

    def where_exists(self, query, boolean="and", negate=False):
        """
        Add an exists clause to the query.

        :param query: The exists query
        :type query: Builder or QueryBuilder

        :type boolean: str

        :type negate: bool

        :rtype: Builder
        """
        if isinstance(query, Builder):
            query = query.get_query()

        self.get_query().where_exists(query, boolean, negate)

        return self

    def or_where_exists(self, query, negate=False):
        """
        Add an or exists clause to the query.

        :param query: The exists query
        :type query: Builder or QueryBuilder

        :type negate: bool

        :rtype: Builder
        """
        return self.where_exists(query, "or", negate)

    def where_not_exists(self, query, boolean="and"):
        """
        Add a where not exists clause to the query.

        :param query: The exists query
        :type query: Builder or QueryBuilder

        :type boolean: str

        :rtype: Builder
        """
        return self.where_exists(query, boolean, True)

    def or_where_not_exists(self, query):
        """
        Add a or where not exists clause to the query.

        :param query: The exists query
        :type query: Builder or QueryBuilder

        :rtype: Builder
        """
        return self.or_where_exists(query, True)

    def has(self, relation, operator=">=", count=1, boolean="and", extra=None):
        """
        Add a relationship count condition to the query.

        :param relation: The relation to count
        :type relation: str

        :param operator: The operator
        :type operator: str

        :param count: The count
        :type count: int

        :param boolean: The boolean value
        :type boolean: str

        :param extra: The extra query
        :type extra: Builder or callable

        :type: Builder
        """
        if relation.find(".") >= 0:
            return self._has_nested(relation, operator, count, boolean, extra)

        relation = self._get_has_relation_query(relation)

        query = relation.get_relation_count_query(
            relation.get_related().new_query(), self
        )

        # TODO: extra query
        if extra:
            if callable(extra):
                extra(query)

        return self._add_has_where(
            query.apply_scopes(), relation, operator, count, boolean
        )

    def _has_nested(self, relations, operator=">=", count=1, boolean="and", extra=None):
        """
        Add nested relationship count conditions to the query.

        :param relations: nested relations
        :type relations: str

        :param operator: The operator
        :type operator: str

        :param count: The count
        :type count: int

        :param boolean: The boolean value
        :type boolean: str

        :param extra: The extra query
        :type extra: Builder or callable

        :rtype: Builder
        """
        relations = relations.split(".")

        def closure(q):
            if len(relations) > 1:
                q.where_has(relations.pop(0), closure)
            else:
                q.has(relations.pop(0), operator, count, boolean, extra)

        return self.where_has(relations.pop(0), closure)

    def doesnt_have(self, relation, boolean="and", extra=None):
        """
        Add a relationship count to the query.

        :param relation: The relation to count
        :type relation: str

        :param boolean: The boolean value
        :type boolean: str

        :param extra: The extra query
        :type extra: Builder or callable

        :rtype: Builder
        """
        return self.has(relation, "<", 1, boolean, extra)

    def where_has(self, relation, extra, operator=">=", count=1):
        """
        Add a relationship count condition to the query with where clauses.

        :param relation: The relation to count
        :type relation: str

        :param extra: The extra query
        :type extra: Builder or callable

        :param operator: The operator
        :type operator: str

        :param count: The count
        :type count: int

        :rtype: Builder
        """
        return self.has(relation, operator, count, "and", extra)

    def where_doesnt_have(self, relation, extra=None):
        """
        Add a relationship count condition to the query with where clauses.

        :param relation: The relation to count
        :type relation: str

        :param extra: The extra query
        :type extra: Builder or callable

        :rtype: Builder
        """
        return self.doesnt_have(relation, "and", extra)

    def or_has(self, relation, operator=">=", count=1):
        """
        Add a relationship count condition to the query with an "or".

        :param relation: The relation to count
        :type relation: str

        :param operator: The operator
        :type operator: str

        :param count: The count
        :type count: int

        :rtype: Builder
        """
        return self.has(relation, operator, count, "or")

    def or_where_has(self, relation, extra, operator=">=", count=1):
        """
        Add a relationship count condition to the query with where clauses and an "or".

        :param relation: The relation to count
        :type relation: str

        :param extra: The extra query
        :type extra: Builder or callable

        :param operator: The operator
        :type operator: str

        :param count: The count
        :type count: int

        :rtype: Builder
        """
        return self.has(relation, operator, count, "or", extra)

    def _add_has_where(self, has_query, relation, operator, count, boolean):
        """
        Add the "has" condition where clause to the query.

        :param has_query: The has query
        :type has_query: Builder

        :param relation: The relation to count
        :type relation: orator.orm.relations.Relation

        :param operator: The operator
        :type operator: str

        :param count: The count
        :type count: int

        :param boolean: The boolean value
        :type boolean: str

        :rtype: Builder
        """
        self._merge_model_defined_relation_wheres_to_has_query(has_query, relation)

        if isinstance(count, basestring) and count.isdigit():
            count = QueryExpression(count)

        return self.where(
            QueryExpression("(%s)" % has_query.to_sql()), operator, count, boolean
        )

    def _merge_model_defined_relation_wheres_to_has_query(self, has_query, relation):
        """
        Merge the "wheres" from a relation query to a has query.

        :param has_query: The has query
        :type has_query: Builder

        :param relation: The relation to count
        :type relation: orator.orm.relations.Relation
        """
        relation_query = relation.get_base_query()

        has_query.merge_wheres(relation_query.wheres, relation_query.get_bindings())

        self._query.add_binding(has_query.get_query().get_bindings(), "where")

    def _get_has_relation_query(self, relation):
        """
        Get the "has" relation base query

        :type relation: str

        :rtype: Builder
        """
        from .relations import Relation

        with Relation.no_constraints(True):
            return getattr(self.get_model(), relation)()

    def with_(self, *relations):
        """
        Set the relationships that should be eager loaded.

        :return: The current Builder instance
        :rtype: Builder
        """
        if not relations:
            return self

        eagers = self._parse_with_relations(list(relations))

        self._eager_load.update(eagers)

        return self

    def _parse_with_relations(self, relations):
        """
        Parse a list of relations into individuals.

        :param relations: The relation to parse
        :type relations: list

        :rtype: dict
        """
        results = {}

        for relation in relations:
            if isinstance(relation, dict):
                for name, constraints in relation.items():
                    results = self._parse_nested_with(name, results)

                    results[name] = constraints

                continue
            else:
                name = relation
                constraints = self.__class__(self.get_query().new_query())

            results = self._parse_nested_with(name, results)

            results[name] = constraints

        return results

    def _parse_nested_with(self, name, results):
        """
        Parse the nested relationship in a relation.

        :param name: The name of the relationship
        :type name: str

        :type results: dict

        :rtype: dict
        """
        progress = []

        for segment in name.split("."):
            progress.append(segment)

            last = ".".join(progress)
            if last not in results:
                results[last] = self.__class__(self.get_query().new_query())

        return results

    def _call_scope(self, scope, *args, **kwargs):
        """
        Call the given model scope.

        :param scope: The scope to call
        :type scope: str
        """
        query = self.get_query()

        # We will keep track of how many wheres are on the query before running the
        # scope so that we can properly group the added scope constraints in the
        # query as their own isolated nested where statement and avoid issues.
        original_where_count = len(query.wheres)

        result = getattr(self._model, scope)(self, *args, **kwargs)

        if self._should_nest_wheres_for_scope(query, original_where_count):
            self._nest_wheres_for_scope(
                query, [0, original_where_count, len(query.wheres)]
            )

        return result or self

    def apply_scopes(self):
        """
        Get the underlying query builder instance with applied global scopes.

        :type: Builder
        """
        if not self._scopes:
            return self

        builder = copy.copy(self)

        query = builder.get_query()

        # We will keep track of how many wheres are on the query before running the
        # scope so that we can properly group the added scope constraints in the
        # query as their own isolated nested where statement and avoid issues.
        original_where_count = len(query.wheres)

        where_counts = [0, original_where_count]

        for scope in self._scopes.values():
            self._apply_scope(scope, builder)

            # Again, we will keep track of the count each time we add where clauses so that
            # we will properly isolate each set of scope constraints inside of their own
            # nested where clause to avoid any conflicts or issues with logical order.
            where_counts.append(len(query.wheres))

        if self._should_nest_wheres_for_scope(query, original_where_count):
            self._nest_wheres_for_scope(query, Collection(where_counts).unique().all())

        return builder

    def _apply_scope(self, scope, builder):
        """
        Apply a single scope on the given builder instance.

        :param scope: The scope to apply
        :type scope: callable or Scope

        :param builder: The builder to apply the scope to
        :type builder: Builder
        """
        if callable(scope):
            scope(builder)
        elif isinstance(scope, Scope):
            scope.apply(builder, self.get_model())

    def _should_nest_wheres_for_scope(self, query, original_where_count):
        """
        Determine if the scope added after the given offset should be nested.

        :type query: QueryBuilder
        :type original_where_count: int

        :rtype: bool
        """
        return original_where_count and len(query.wheres) > original_where_count

    def _nest_wheres_for_scope(self, query, where_counts):
        """
        Nest where conditions of the builder and each global scope.

        :type query: QueryBuilder
        :type where_counts: list
        """
        # Here, we totally remove all of the where clauses since we are going to
        # rebuild them as nested queries by slicing the groups of wheres into
        # their own sections. This is to prevent any confusing logic order.
        wheres = query.wheres

        query.wheres = []

        # We will take the first offset (typically 0) of where clauses and start
        # slicing out every scope's where clauses into their own nested where
        # groups for improved isolation of every scope's added constraints.
        previous_count = where_counts.pop(0)

        for where_count in where_counts:
            query.wheres.append(
                self._slice_where_conditions(
                    wheres, previous_count, where_count - previous_count
                )
            )

            previous_count = where_count

    def _slice_where_conditions(self, wheres, offset, length):
        """
        Create a where list with sliced where conditions.

        :type wheres: list
        :type offset: int
        :type length: int

        :rtype: list
        """
        where_group = self.get_query().for_nested_where()
        where_group.wheres = wheres[offset : (offset + length)]

        return {"type": "nested", "query": where_group, "boolean": "and"}

    def get_query(self):
        """
        Get the underlying query instance.

        :rtype: QueryBuilder
        """
        return self._query

    def to_base(self):
        """
        Get a base query builder instance.

        :rtype: QueryBuilder
        """
        return self.apply_scopes().get_query()

    def set_query(self, query):
        """
        Set the underlying query instance.

        :param query: A QueryBuilder instance
        :type query: QueryBuilder
        """
        self._query = query

    def get_eager_loads(self):
        """
        Get the relationships being eager loaded.

        :rtype: dict
        """
        return self._eager_load

    def set_eager_loads(self, eager_load):
        """
        Sets the relationships to eager load.

        :type eager_load: dict

        :rtype: Builder
        """
        self._eager_load = eager_load

        return self

    def get_model(self):
        """
        Get the model instance of the model being queried

        :rtype: orator.Model
        """
        return self._model

    def set_model(self, model):
        """
        Set a model instance for the model being queried.

        :param model: The model instance
        :type model: orator.orm.Model

        :return: The current Builder instance
        :rtype: Builder
        """
        self._model = model

        self._query.from_(model.get_table())

        return self

    def macro(self, name, callback):
        """
        Extend the builder with the given callback.

        :param name: The extension name
        :type name: str

        :param callback: The callback
        :type callback: callable
        """
        self._macros[name] = callback

    def get_macro(self, name):
        """
        Get the given macro by name

        :param name: The macro name
        :type name: str
        :return:
        """
        return self._macros.get(name)

    def __dynamic(self, method):
        from .utils import scope

        scope_method = "scope_%s" % method
        is_scope = False
        is_macro = False

        # New scope definition check
        if hasattr(self._model, method) and isinstance(
            getattr(self._model, method), scope
        ):
            is_scope = True
            attribute = getattr(self._model, method)
            scope_method = method
        # Old scope definition check
        elif hasattr(self._model, scope_method):
            is_scope = True
            attribute = getattr(self._model, scope_method)
        elif method in self._macros:
            is_macro = True
            attribute = self._macros[method]
        else:
            if method in self._passthru:
                attribute = getattr(self.apply_scopes().get_query(), method)
            else:
                attribute = getattr(self._query, method)

        def call(*args, **kwargs):
            if is_scope:
                return self._call_scope(scope_method, *args, **kwargs)
            if is_macro:
                return attribute(self, *args, **kwargs)

            result = attribute(*args, **kwargs)

            if method in self._passthru:
                return result
            else:
                return self

        if not callable(attribute):
            return attribute

        return call

    def __getattr__(self, item, *args):
        return self.__dynamic(item)

    def __copy__(self):
        new = self.__class__(copy.copy(self._query))
        new.set_model(self._model)

        return new



from setuptools import setup, find_packages

setup(
    name="orbit-orator",
    version="0.9.9.5",
    description="A simple ORM for Python",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Maxiaoyu",
    author_email="maxiaoyu@example.com",
    url="https://github.com/maxiaoyu/orator",
    packages=find_packages(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    python_requires=">=3.8",
) 


# -*- coding: utf-8 -*-

from pendulum import DateTime, Date

try:
    import sqlite3

    from sqlite3 import register_adapter

    register_adapter(DateTime, lambda val: val.isoformat(" "))
    register_adapter(Date, lambda val: val.isoformat())
except ImportError:
    sqlite3 = None

from ..dbal.platforms import SQLitePlatform
from ..utils.helpers import serialize
from .connector import Connector


class DictCursor(dict):
    def __init__(self, cursor, row):
        self.dict = {}
        self.cursor = cursor

        for idx, col in enumerate(cursor.description):
            self.dict[col[0]] = row[idx]

        super(DictCursor, self).__init__(self.dict)

    def __getattr__(self, item):
        try:
            return self[item]
        except KeyError:
            return getattr(self.cursor, item)

    def serialize(self):
        return serialize(self)


class SQLiteConnector(Connector):

    RESERVED_KEYWORDS = [
        "log_queries",
        "driver",
        "prefix",
        "name",
        "foreign_keys",
        "use_qmark",
    ]

    def _do_connect(self, config):
        connection = self.get_api().connect(**self.get_config(config))
        connection.isolation_level = None
        connection.row_factory = DictCursor

        # We activate foreign keys support by default
        if config.get("foreign_keys", True):
            connection.execute("PRAGMA foreign_keys = ON")

        return connection

    def get_api(self):
        return sqlite3

    @property
    def isolation_level(self):
        return self._connection.isolation_level

    @isolation_level.setter
    def isolation_level(self, value):
        self._connection.isolation_level = value

    def get_dbal_platform(self):
        return SQLitePlatform()

    def is_version_aware(self):
        return False

    def get_server_version(self):
        sql = "select sqlite_version() AS sqlite_version"

        rows = self._connection.execute(sql).fetchall()
        version = rows[0]["sqlite_version"]

        return tuple(version.split(".")[:3] + [""])



# -*- coding: utf-8 -*-

from .belongs_to_many import BelongsToMany


class MorphToMany(BelongsToMany):
    def __init__(
        self,
        query,
        parent,
        name,
        table,
        foreign_key,
        other_key,
        relation_name=None,
        inverse=False,
    ):
        """
        :param query: A Builder instance
        :type query: elquent.orm.Builder

        :param parent: The parent model
        :type parent: Model

        :param table: The pivot table
        :type table: str

        :param foreign_key: The foreign key
        :type foreign_key: str

        :param other_key: The other key
        :type other_key: str

        :param relation_name: The relation name
        :type relation_name: str

        :type inverse: bool
        """
        self._name = name
        self._inverse = inverse
        self._morph_type = name + "_type"
        self._morph_name = (
            query.get_model().get_morph_name() if inverse else parent.get_morph_name()
        )

        super(MorphToMany, self).__init__(
            query, parent, table, foreign_key, other_key, relation_name
        )

    def _set_where(self):
        """
        Set the where clause for the relation query.

        :return: self
        :rtype: BelongsToMany
        """
        super(MorphToMany, self)._set_where()

        self._query.where("%s.%s" % (self._table, self._morph_type), self._morph_name)

    def get_relation_count_query(self, query, parent):
        """
        Add the constraints for a relationship count query.

        :type query: orator.orm.Builder
        :type parent: orator.orm.Builder

        :rtype: orator.orm.Builder
        """
        query = super(MorphToMany, self).get_relation_count_query(query, parent)

        return query.where("%s.%s" % (self._table, self._morph_type), self._morph_name)

    def add_eager_constraints(self, models):
        """
        Set the constraints for an eager load of the relation.

        :type models: list
        """
        super(MorphToMany, self).add_eager_constraints(models)

        self._query.where("%s.%s" % (self._table, self._morph_type), self._morph_name)

    def _create_attach_record(self, id, timed):
        """
        Create a new pivot attachement record.
        """
        record = super(MorphToMany, self)._create_attach_record(id, timed)

        record[self._morph_type] = self._morph_name

        return record

    def _new_pivot_query(self):
        """
        Create a new query builder for the pivot table.

        :rtype: orator.orm.Builder
        """
        query = super(MorphToMany, self)._new_pivot_query()

        return query.where(self._morph_type, self._morph_name)

    def new_pivot(self, attributes=None, exists=False):
        """
        Create a new pivot model instance.
        """
        from .morph_pivot import MorphPivot

        pivot = MorphPivot(self._parent, attributes, self._table, exists)

        pivot.set_pivot_keys(self._foreign_key, self._other_key).set_morph_type(
            self._morph_type
        ).set_morph_name(self._morph_name)

        return pivot

    def get_morph_type(self):
        return self._morph_type

    def get_morph_name(self):
        return self._morph_name

    def _new_instance(self, model):
        return MorphToMany(
            self.new_query(),
            model,
            self._name,
            self._table,
            self._foreign_key,
            self._other_key,
            self._relation_name,
            self._inverse,
        )
